{-# LANGUAGE OverloadedStrings #-}
module MyLib (defaultMain) where

import Control.Exception (bracket)
import Control.Lens
import Control.Monad (void)
import qualified Data.Aeson as A
import qualified Data.ByteString.Lazy as LB
import Data.Config
import Data.Foldable (for_)
import Data.Functor.Contravariant ((>$<))
import Data.Text (Text)
import qualified Data.Text.Strict.Lens as T
import Network.AMQP
import Network.Reddit (publishEntries)
import Network.Wreq.Session (newSession)
import Publish (Publish(..))


amqpPublisher :: Channel -> Text -> Publish IO LB.ByteString
amqpPublisher channel exchange = Publish $ \lbs ->
  void $ publishMsg channel exchange routingKey (message lbs)
  where
    routingKey = "doesn't matter on fanout"
    message lbs = newMsg
      { msgBody = lbs
      , msgDeliveryMode = Just Persistent
      }

defaultMain :: IO ()
defaultMain = do
  conf <- readConfig "./config.dhall"
  let connect = openConnection
                  (conf ^. amqp . host . T.unpacked)
                  (conf ^. amqp . vhost)
                  (conf ^. amqp . username)
                  (conf ^. amqp . password)
  bracket connect closeConnection $ \conn -> do
    putStrLn "Hello"
    chan <- openChannel conn
    declareExchange chan newExchange { exchangeName = "reddit_posts", exchangeType = "fanout" }
    sess <- newSession
    let encoder = amqpPublisher chan "reddit_posts"
    for_ (conf ^. fetchers) $ \fetcher -> do
      print fetcher
      publishEntries (A.encode >$< encoder) sess fetcher
